#include <stdio.h>
#include <stdlib.h> /* to get the random number library */
#include <string.h>
#include "gfd.h"

#define MAX_TIMERS 1000

static inline void timer_get( long long *ptime )
{
#  if (__GNUC__ > 2) || ((__GNUC__ == 2) &&(__GNUC_MINOR__ >= 95))
    __asm__ __volatile__ ("rdtsc" : "=A"(*ptime));
#  else
    __asm__ __volatile__ (
        ".byte 0x0f; .byte 0x31"
        : "=A" (*ptime)
        : : "eax", "edx" );
#  endif  /* (__GNUC__ > 2) || ((__GNUC__ == 2) &&(__GNUC_MINOR__ > 95)) */
}

static inline void timer_add( long long *ps, long long *pt )
{
   *ps = *ps + *pt;
}

static inline long long timer_diff( long long te, long long ts )
{
   return (te - ts);
}

static inline double timer_ms( long long t )
{
  /* as reported in /proc/cpuinfo */
  /* torc5.cs.utk.du  : Pentium III				CPU  547.186 MHz */
  /* earth.cs.utk.edu : Intel(R) Xeon(TM)		CPU 2392.307 MHz */
	
   /* double freq = 547.186; */
   double freq = 2392.307 ;
   return (((double)(t)) / freq);
}

int main(int argc, char **argv){

   unsigned int x;
   unsigned int* X1;
   unsigned int* X2;
   unsigned int* X ; 
   unsigned int* s1;
   unsigned int* s2;
   unsigned int* s;
   unsigned int* s_ ; 

   unsigned int* gflog ; 
   unsigned int* gfilog ; 

   int w, wg, m ;
   int nx,nx1,nx2,nx3,nx4;	
   int i,k, index;
   int seed ;
   int seed_;
   int iter, maxit;

   /* timers */
   long long timer_sig_gfd, timer_conc_gfd, timer_uni_gfd;
   long long timer_sig_crc, timer_conc_crc;
   long long timer_temp1, timer_temp2;

   wg  =   48; /* 13 or 48 predefined datatypes  */    /* 13 < 16=2^4 and 48 < 64=2^6 */
   w   =    8; /* length of a hash function      */   /* 32 = 2^5 */
   m   =    2; /* number of hash functions       */
   nx  =  100;  
   nx1 =    0;  
   nx2 =    0;  
   nx3 =    0;  
   nx4 =    0;  
   seed = 1;
   maxit = 10;

   for( i = 1; i < argc; i++ ) {
      if( strcmp( argv[i], "-nx" ) == 0 ) { 
         nx  = atoi(argv[i+1]);
         i++;
      }
      if( strcmp( argv[i], "-nx1" ) == 0 ) { 
         nx1  = atoi(argv[i+1]);
         i++;
      }
      if( strcmp( argv[i], "-nx2" ) == 0 ) { 
         nx2  = atoi(argv[i+1]);
         i++;
      }
      if( strcmp( argv[i], "-w" ) == 0 ) { 
         w = atoi(argv[i+1]);
         i++;
      }
      if( strcmp( argv[i], "-m" ) == 0 ) {
         m = atoi(argv[i+1]);
         i++;
      }
      if( strcmp( argv[i], "-wg" ) == 0 ) {
         wg = atoi(argv[i+1]);
         i++;
      }
      if( strcmp( argv[i], "-iter" ) == 0 ) {
         maxit = atoi(argv[i+1]);
         i++;
      }
      if( strcmp( argv[i], "-seed" ) == 0 ) {
         seed = atoi(argv[i+1]);
         i++;
      }
   }

   if (nx1==0){
      nx1 = nx/2  ;
   }
   if (nx2==0){
      nx2 = nx1 ;
   }

   if (w*m>32){
      printf("BUG\n");
      printf("at this point we think to store the hash function on a unsigned int of 32 bits\n");
      printf("thus m*n shall be smaller than 32 - this is not your case\n");
      exit(1);
   }

   X1 = (unsigned int *)malloc(nx1*sizeof(unsigned int)) ;
   if (X1==NULL){ printf("error of memory allocation\n"); exit(0); }
   s1 = (unsigned int *)malloc(2*sizeof(unsigned int)) ;
   if (s1==NULL){ printf("error of memory allocation\n"); exit(0); }

   X2 = (unsigned int *)malloc(nx2*sizeof(unsigned int)) ;
   if (X2==NULL){ printf("error of memory allocation\n"); exit(0); }
   s2 = (unsigned int *)malloc(2*sizeof(unsigned int)) ;
   if (s2==NULL){ printf("error of memory allocation\n"); exit(0); }

   nx = nx1+nx2;

   X  = (unsigned int *)malloc(nx*sizeof(unsigned int)) ;
   if (X ==NULL){ printf("error of memory allocation\n"); exit(0); }
   s  = (unsigned int *)malloc(2*sizeof(unsigned int)) ;
   if (s ==NULL){ printf("error of memory allocation\n"); exit(0); }
   s_ = (unsigned int *)malloc(2*sizeof(unsigned int)) ;
   if (s_==NULL){ printf("error of memory allocation\n"); exit(0); }

   gfd_gfd_setuptabl( w , &gflog , &gfilog ) ;

   for (iter=0;iter<maxit;iter++){

      seed_ = seed + iter;	
      srand(seed_);	

      printf( "---------------------------------------------------------------------------\n");
      printf( " ( nx1 = %3d , nx2 = %3d , wg = %3d , seed = %3d ) --  ( w = %2d , m = %1d ) :\n",
              nx1,nx2,wg,seed_,w,m );
	
      for (i=0;i<nx1;i++){
         while (wg <= (X1[i] = rand() / (RAND_MAX/wg)) );
         X[i]    = X1[i];
      }
      for (i=0;i<nx2;i++){
         while (wg <= (X2[i] = rand() / (RAND_MAX/wg)) );
         X[nx1+i] = X2[i];
      }

      gfd_gfd_signature( s1 , X1 , nx1 , w , m , gflog , gfilog ) ;
      gfd_gfd_signature( s2 , X2 , nx2 , w , m , gflog , gfilog ) ;

      timer_sig_gfd = 0;
      for( index = 0; index < MAX_TIMERS; index++ ) {
         timer_get( &timer_temp1 );
         gfd_gfd_signature( s  , X  , nx  , w , m , gflog , gfilog ) ;
         timer_get( &timer_temp2 );
         timer_sig_gfd += timer_diff( timer_temp2, timer_temp1 );
      }

      timer_conc_gfd = 0;
      for( index = 0; index < MAX_TIMERS; index++ ) {
		 timer_get( &timer_temp1 );
		 gfd_gfd_concatsig( s_ , s1 , s2  , w , m , gflog , gfilog ) ;
         timer_get( &timer_temp2 );
         timer_conc_gfd += timer_diff( timer_temp2, timer_temp1 );
      }

      printf("\n");
      /*printf("%10lld\n", timer_sig_gfd / MAX_TIMERS);*/
      printf("Timer for gfd_signature %8.2f microseconds \tnx =%d\n", timer_ms(timer_sig_gfd / MAX_TIMERS),nx );
      /*printf("%10lld\n", timer_conc_gfd / MAX_TIMERS);*/
      printf("Timer for gfd_concat    %8.2f microseconds \tnx2=%d\n", timer_ms(timer_conc_gfd / MAX_TIMERS),nx2 );
      printf("Check if concat ok :  %8X   %8X\n",s[0],s_[0]);
      if (!(s[0]==s_[0])){printf(" BUG"); exit(1);}
      printf("\n");
	

      {
         unsigned short crc = 0xFFFF;
         unsigned short temp;
         unsigned short quick;
         for( k = 0; k < nx1; k++ ) {
            unsigned char byte = X1[k]+'0' ;
        
            temp = (crc >> 8) ^ byte;
            crc <<= 8;
            quick = temp ^ (temp >> 4);
            crc ^= quick;
            quick <<= 5;
            crc ^= quick;
            quick <<= 7;
            crc ^= quick;
         }
         *s1 = crc ;
	  }	
      {
         unsigned short crc = 0xFFFF;
         unsigned short temp;
         unsigned short quick;
         for( k = 0; k < nx2; k++ ) {
            unsigned char byte = X2[k]+'0' ;
        
            temp = (crc >> 8) ^ byte;
            crc <<= 8;
            quick = temp ^ (temp >> 4);
            crc ^= quick;
            quick <<= 5;
            crc ^= quick;
            quick <<= 7;
            crc ^= quick;
         }
         *s2 = crc ;
	  }	

      timer_sig_crc = 0;
      for( index = 0; index < MAX_TIMERS; index++ ) {
         timer_get( &timer_temp1 );

         {
            unsigned short crc = 0xFFFF;
            unsigned short temp;
            unsigned short quick;
            for( k = 0; k < nx; k++ ) {
               unsigned char byte = X[k]+'0' ;
        
               temp = (crc >> 8) ^ byte;
               crc <<= 8;
               quick = temp ^ (temp >> 4);
               crc ^= quick;
               quick <<= 5;
               crc ^= quick;
               quick <<= 7;
               crc ^= quick;
            }
            *s = crc ;

            timer_get( &timer_temp2 );
            timer_sig_crc += timer_diff( timer_temp2, timer_temp1 );
         }
      }

      timer_conc_crc = 0;
      for( index = 0; index < MAX_TIMERS; index++ ) {
         timer_get( &timer_temp1 );

		 {
         unsigned short crc = *s1;
         unsigned short temp;
         unsigned short quick;
         for( k = 0; k < nx2; k++ ) {
            unsigned char byte = X2[k]+'0' ;

            temp = (crc >> 8) ^ byte;
            crc <<= 8;
            quick = temp ^ (temp >> 4);
            crc ^= quick;
            quick <<= 5;
            crc ^= quick;
            quick <<= 7;
            crc ^= quick;
         }
         *s_ = crc ;
		 }

         timer_get( &timer_temp2 );
         timer_conc_crc += timer_diff( timer_temp2, timer_temp1 );
      }
   
      /*printf("%10lld\n", timer_sig_crc / MAX_TIMERS);*/
      printf("Timer for crc_signature %8.2f microseconds \tnx =%d\n", timer_ms(timer_sig_crc / MAX_TIMERS),nx );
      /*printf("%10lld\n", timer_conc_crc / MAX_TIMERS);*/
      printf("Timer for crc_concat    %8.2f microseconds \tnx2=%d\n", timer_ms(timer_conc_crc / MAX_TIMERS),nx2 );
      printf("Check if concat ok :  %8X   %8X\n",s[0],s_[0]);
      if (!(s[0]==s_[0])){printf(" BUG"); exit(1);}
      printf("\n");

      seed_ = seed + iter;	
      srand(seed_);	
      while (wg <= (x = rand() / (RAND_MAX/wg)) );
      for (i=0;i<nx;i++){
         X[i] = x;
      }

      /*printf(" ( nx = %2d , x = %d ,  wg = %3d , seed = %3d ) --  ( w = %2d , m = %1d ) :",nx,x,wg,seed_,w,m);*/

      gfd_gfd_signature( s , X , nx , w , m , gflog , gfilog ) ;

      timer_uni_gfd = 0;
      for( index = 0; index < MAX_TIMERS; index++ ) {
		 timer_get( &timer_temp1 );
         gfd_gfd_unisignat( s_ , x , nx , w , m , gflog , gfilog ) ;
         timer_get( &timer_temp2 );
         timer_uni_gfd += timer_diff( timer_temp2, timer_temp1 );
      }

      printf("Timer for gfd_uni       %8.2f microseconds \tnx =%d\n", timer_ms(timer_uni_gfd / MAX_TIMERS),nx );
      printf("Check if uni    ok :  %8X   %8X\n",s[0],s_[0]);
      if (!(s[0]==s_[0])){printf(" BUG"); exit(1);}
   }	
   printf( "---------------------------------------------------------------------------\n");

   free(X);
   free(X1);
   free(X2);
   free(s1);
   free(s2);
   free(s);
   free(s_);
   return 0;

}

